/*
	VRay/Max related code mostly direct copy of VRaySphere
	
	META part is adaptation of Keenan Crane's META "Ray Tracing Quaternion META Sets on the GPU"
	
	The rest by Jerry Ylilammi	
*/


#include "max.h"
#include "BerconMetaball.h"

using namespace VR;

/*Ireal fieldFunction(Ireal size, Ireal radius) {
	radius = sqrt(radius);
	if (radius <= (size/3.0f))
		return ((1.f - 3.f * radius * radius) / (size * size));			
	Ireal temp = (1.f - radius / size);
	return 3.f / 2.f * temp * temp;
}*/

// All take SQUARED size and SQUARED radius

Ireal fieldFunctionSimple(Ireal size, Ireal radius) {
	Ireal temp = (1. - size / radius);
	return temp * temp;
}

Ireal fieldFunctionBlinn(Ireal size, Ireal radius) {
	return exp(-(size / radius));
}

Ireal fieldFunctionWyvill(Ireal size, Ireal radius) {
	Ireal temp = radius / size;	
	Ireal temp2 = temp * temp; 		
	return 1. + 17. / 9. * temp2 - 4.0 / 9. * temp2 * temp - 22. / 9. * temp;	
}

Ireal fieldFunctionRenderMan(Ireal size, Ireal radius) {
	Ireal temp = radius / size;	
	Ireal temp2 = temp * temp; 			
	return 1. - 3.* temp + 3.* temp2 - temp*temp2;	
}

	/*if (rayp.length() > delta) {
		float r = delta;
		float ac=raydir*raydir;
		float bc=2.0f*rayp*raydir;
		float cc=rayp*rayp-r*r;
		float D=bc*bc-4.0f*ac*cc;
		if (D<0.0f) return false;
		D=sqrtf(D);
		Ireal s0 =(-bc-D)/(2.0f*ac);
		Ireal s1 =(-bc+D)/(2.0f*ac);
		rO = rayp + raydir * (s0>s1?s1:s0);
	} else {
		rO = rayp;
	}*/

/*
	Ray intersection

	Rays startting from outside are handled differently to those starting
	from inside because we this allows more optimizations.
*/
int BerconMetaball::intersect(RSRay &ray) {	
	this->stat_RaysCast++; // STATISTICS		

	Vector rO = ray.p;
	real dirLength = ray.dir.length();
	Vector rD = ray.dir;

	this->stat_avgMint += ray.cmint;
	this->stat_avgDirLength += dirLength;
	
	Ireal stepLength = this->stepLength / dirLength;
	Ireal maxError = this->maxError / dirLength;
			
	Ireal dist = ray.cmint;
		  
	if (ray.skipTag == this)
		dist += maxError * 2.;

	 // Infinite range, not cmint/cmaxt because it caused black "sliceplane" along x or y or x axis
	IRay tRay = IRay(rO, rD);
	//IRay tRay = IRay(rO, rD, mint, maxt);
	real f = 0.f;			
	bvh->getFieldValue(blobs, rO+rD*dist, blobSize, blobSize2, function, &f);
	
	bool insideField = f > blobThreshold;	

	BVHRay bRay = BVHRay();
	bvh->traceRay(&tRay, &bRay);	

	if (bRay.first == NULL) // No intersection with the bounding volume
		return false;

	BVHStep* cur;
	BVHStep* prev;
	BVHStep* step = bRay.first;		

	/*cur = step; // STATISTICS
	if (insideField) {
		stat_RaysIn++;
		while (cur != NULL) {
			cur = cur->next;			
			stat_bRayDepthIn++;			
		}
	} else {
		stat_RaysOut++;
		while (cur != NULL) {
			cur = cur->next;			
			stat_bRayDepthOut++;
		} 
	}*/ // END

	/*dist = stepLength; //maxError*2.;
	while (dist<ray.cmaxt){
		real f = 0.f;
		bvh->getFieldValue(blobs, rO+rD*dist, blobSize, blobSize2, &fieldFunctionRenderMan, &f);
		if (testThreshold(f, insideField))
			break;
		dist += stepLength;
	}*/
	
	Ireal distLate;	
	if (insideField) { // Inside		
		//dist = maxError_2;
		distLate = dist;
		while (step != NULL) {		
			f = 0.f;
			cur = step;
			prev = NULL;		
			while (cur != NULL && dist > cur->min) {
				// Optimize the ray BVHRay on the fly, however do it one step late because
				// we might need to use the step information when searching the exact hit.				
				if (distLate > cur->max) {
					if (cur == step)
						step = step->next;
					else {						
						prev->next = cur->next;
						cur->deleteThisOnly();
						cur = prev;
					}
					if (step == NULL) return false;
					cur = cur->next;
					continue;
				}
				if (dist < cur->max) {
					cur->cont->getFieldValue(blobs, rO+rD*dist, blobSize, blobSize2, function, &f);				
					if (testThreshold(f, false))
						break;
				}
				prev = cur;
				cur = cur->next;
			}
			if (testThreshold(f, insideField))
				break;
			distLate = dist;
			dist += stepLength;
		}
	} else { // Outside
		//dist = step->min;
		distLate = dist;
		while (step != NULL) {		
			f = 0.f;
			cur = step;
			prev = NULL;		
			while (cur != NULL && dist > cur->min) {
				// Optimize the ray BVHRay on the fly, however do it one step late because
				// we might need to use the step information when searching the exact hit.				
				if (distLate > cur->max) {
					if (cur == step)
						step = step->next;
					else {						
						prev->next = cur->next;
						cur->deleteThisOnly();
						cur = prev;
					}
					if (step == NULL) return false;
					cur = cur->next;
					continue;
				}
				if (dist < cur->max)
					cur->cont->getFieldValue(blobs, rO+rD*dist, blobSize, blobSize2, function, &f);				
				prev = cur;
				cur = cur->next;
			}
			if (testThreshold(f, insideField))
				break;
			if (dist < step->min) {
				dist = step->min;
				distLate = dist;
			} else {
				distLate = dist;
				dist += stepLength;
			}
		}
	}
	
	if (step == NULL) return false; // No hit with the field

	// Confirmed hit, now we need to achieved desired maximum error	
	Ireal error = stepLength / 2.;
	dist -= error;		
	while (error > maxError) {
		f = 0.f;
		cur = step;
		prev = NULL;		
		while (cur != NULL && dist > cur->min) {			
			if (dist < cur->max)
				cur->cont->getFieldValue(blobs, rO+rD*dist, blobSize, blobSize2, function, &f);				
			prev = cur;
			cur = cur->next;
		}
		error /= 2.;		
		if (testThreshold(f, insideField))
			dist -= error;
		else
			dist += error;
	}
	
	if (dist>ray.is.t) return false;
	if (dist<=ray.cmint) return false;	
	if (dist>=ray.cmaxt) return false;

	MyCacheStruct cacheStruct;
	cacheStruct.rO = ray.p + ray.dir * dist;

	ray.is.t = dist;
	ray.is.primitive = (GenericPrimitive*) this;
	ray.is.skipTag = this;
	
	raycache->putCache(*(VR::VRayContext*) ray.rayparams, cacheStruct);

	return true;
}

Vector BerconMetaball::getGNormal(RSRay &ray) {
	MyCacheStruct cacheStruct;
	raycache->getCache(*(VR::VRayContext*) ray.rayparams, cacheStruct);
	Vector N = bvh->getFieldGrad(blobs, cacheStruct.rO, blobSize, blobSize2, function);
	//if (!this->simpleSplit) N = -N;
	//normalFetch.Append(1,&(cacheStruct.rO),500);
	return N;
}

Vector BerconMetaball::getGNormal(RSRayRef& rayref) {
	MyCacheStruct cacheStruct;
	raycache->getCache(*(VR::VRayContext*) rayref.getRayParams(), cacheStruct);	
	Vector N = bvh->getFieldGrad(blobs, cacheStruct.rO, blobSize, blobSize2, function);
	//if (!this->simpleSplit) N = -N;
	//normalFetch.Append(1,&(cacheStruct.rO),500);
	return N;
}

/*int BerconMetaball::intersect(RSRay &ray) {
	MyCacheStruct cacheStruct;

	float r = blobSize;
	// if (!raycache->getCache(*(VR::VRayContext*) ray.rayparams, cacheStruct)) {
	Vector offs=Vector(ray.p) - blobs[0];

		float ac=ray.dir*ray.dir;
		float bc=2.0f*offs*ray.dir;
		float cc=offs*offs-r*r;

		float D=bc*bc-4.0f*ac*cc;
		if (D<0.0f) return false;

		D=sqrtf(D);
		cacheStruct.s0=(-bc-D)/(2.0f*ac);
		cacheStruct.s1=(-bc+D)/(2.0f*ac);

		// Make s0 to be in front of s1 always
		if (cacheStruct.s0>cacheStruct.s1) swap(cacheStruct.s0, cacheStruct.s1);

		// raycache->putCache(*(VR::VRayContext*) ray.rayparams, cacheStruct);
	// }

	Ireal s0=cacheStruct.s0;
	Ireal s1=cacheStruct.s1;

  if (this==ray.skipTag) {
		// Leave only the intersection that is farther from the ray origin
		//if (fabsf(s0-ray.mint)<fabsf(s1-ray.mint)) s0=s1;
		if (fabsf(s0)<fabsf(s1)) s0=s1;
		else s1=s0;
	}

	// If s0 is in front of the segment start, use s1
	// If s1 is in front of the segment start too, then there is no intersection
	if (s0<=ray.cmint) {
		if (s1<=ray.cmint) return false;
		s0=s1;
	}

	// At this point, s0 is potentially the intersection we need
	// If it is after the segment end, or after another intersection of the ray,
	// then report no intersection
	if (s0>=ray.cmaxt || s0>ray.is.t) return false;

	ray.is.t=s0;
	ray.is.primitive=(GenericPrimitive*) this;
	ray.is.skipTag=this;

	cacheStruct.nrm=(Vector(ray.p)+ray.dir*s0)-p;
	raycache->putCache(*(VR::VRayContext*) ray.rayparams, cacheStruct);

	return true;
}

Vector BerconMetaball::getGNormal(RSRay &ray) {
	MyCacheStruct cacheStruct;
	raycache->getCache(*(VR::VRayContext*) ray.rayparams, cacheStruct);

	//Vector ip=Vector(ray.p)+ray.dir*real(ray.is.t);
	// Vector nrm=(ip-p);
	Vector nrm=cacheStruct.nrm;
	return nrm;
}

Vector BerconMetaball::getGNormal(RSRayRef& rayref) 
{
	MyCacheStruct cacheStruct;
	raycache->getCache(*(VR::VRayContext*) rayref.getRayParams(), cacheStruct);

	//Vector ip=rayref.getOrigin() + rayref.getDirection() * real(rayref.getDistance());
	// Vector nrm=(ip-p);
	Vector nrm=cacheStruct.nrm;
	return nrm;
}*/

void BerconMetaball::getBBox(StaticBox &b) {
	b = bb;
}

void BerconMetaball::split(int dim, real middle, StaticBox &bLeft, StaticBox &bRight) {
	bb.split(dim, middle, bLeft, bRight);
	return;

	bLeft.init();
	for (int i=0;i<blobsCount;i++)
		bLeft.addExpanded(blobs[i], (real)blobSize);
	bRight.init();
	for (int i=0;i<blobsCount;i++)
		bRight.addExpanded(blobs[i], (real)blobSize);
	/*
	bLeft.init();
	bRight.init();

	for (int i=0;i<blobsCount;i++) {
		bLeft.addExpanded(blobs[i], blobSize);
		bRight.addExpanded(blobs[i], blobSize);
	}
	*/
	/*
	for (int i=0;i<blobsCount;i++) {
		if (blobs[i][dim] < middle + blobSize)
			bLeft.addExpanded(blobs[i], blobSize);
		if (blobs[i][dim] > middle - blobSize)
			bRight.addExpanded(blobs[i], blobSize);
	}

	if (bLeft.pmax[dim] > middle) bLeft.pmax[dim]=middle;
	if (bRight.pmin[dim] > middle) bRight.pmin[dim]=middle;*/
}

void BerconMetaball::init(
		int maxIter, int maxNorm, float epsilon, float delta,
		float cX, float cY, float cZ, float cW, float sX, float sY, float sZ, float sW, float slice,
		Vector *points, int pointCount,
		int simpleSplit,
		VR::GeometryGenerator *owner, int ownerIndex, MyRayCache *rayc) {
	
	//p = oTW.offs;

	//this->maxIter = maxIter;
	//this->maxNorm = maxNorm;
	//this->epsilon = (double)epsilon; // / (double)1000;
	//this->delta = (double)delta; // / (double)1000;

	switch (maxNorm) {
		default: this->function = &fieldFunctionSimple;	break;
		case 2: this->function = &fieldFunctionBlinn; break;
		case 3: this->function = &fieldFunctionWyvill; break;
		case 4: this->function = &fieldFunctionRenderMan; break;
	}

	this->blobSize = delta;
	this->blobSize2 = delta*delta;
	this->blobThreshold = slice;
	this->blobStrength = cY;
	this->stepLength = epsilon;

	this->maxError = cX;

	this->simpleSplit = simpleSplit;
	
	this->blobs = points;
	this->blobsCount = pointCount;

	bb.init();
	for (int i=0;i<blobsCount;i++)
		bb.addExpanded(blobs[i], (real)blobSize);

	clock_t start, finish; // STATISTICS
	start = clock();
	this->bvh->create(this->blobs, this->blobsCount, (real)this->blobSize, maxIter);
	finish = clock();
	this->stat_bvhBuildTime = (finish - start);//CLOCKS_PER_SEC; // END
	
	//this->c = double4(cX, cY, cZ, cW);	
	//this->sliceP = double4(sX, sY, sZ, sW);	
	this->slice = slice;

	this->owner=owner;
	this->ownerIndex=ownerIndex;
	raycache=rayc;
}