#ifndef __PRIMITIVE_H__
#define __PRIMITIVE_H__

#include "vraybase.h"
#include "vraygeom.h"
#include "rayserver.h"

#include "vrayrenderer.h"
#include "vrayplugins.h"
#include "vraytexutils.h"

#include "BoundingVolumeHierarchy.h"

#define BERCON_METABALL_PRIM 0x10011256

struct MyCacheStruct {	
	VR::Vector rO;
	VR::Ireal s0, s1;
	VR::Vector nrm;
};

typedef VR::RayCache<MyCacheStruct> MyRayCache;

class BerconMetaball: public VR::StaticPrimitive {
	/* Metaball stuff */

private:

	// Parameters	
	VR::Ireal blobSize;	
	VR::Ireal blobThreshold;
	VR::Ireal blobStrength;
	VR::Ireal stepLength;
	VR::Ireal maxError;
	int simpleSplit;	
	//VR::Ireal (*function)(VR::Ireal, VR::Ireal)
	VR::Ireal (*function)(VR::Ireal, VR::Ireal);

	// Generated values
	VR::Ireal blobSize2;	
	VR::StaticBox bb;

	// Metaballs
	VR::Vector *blobs;	
	int blobsCount;
	BoundingVolumeHierarchy* bvh;
	bool testThreshold(float f, bool insideField) {if (insideField) return f < blobThreshold; else return f > blobThreshold;}

	// Statistics
	long stat_RaysCast;
	long stat_RaysSkipped;	
	long stat_RaysOut;
	long stat_RaysIn;
	long stat_HitsOut;
	long stat_HitsIn;
	long stat_bRayDepthOut;
	long stat_bRayDepthIn;
	long stat_bvhBuildTime;

	double stat_avgMint;
	double stat_avgDirLength;

	long stat_insideLevel[10];

	Tab<VR::Vector> normalFetch;
	
public:
	BerconMetaball() {
		bvh = new BoundingVolumeHierarchy();

		stat_RaysCast = 0L;
		stat_RaysSkipped = 0L;		
		stat_RaysOut = 0L;
		stat_RaysIn = 0L;
		stat_HitsOut = 0L;
		stat_HitsIn = 0L;
		stat_bRayDepthOut = 0L;
		stat_bRayDepthIn = 0L;
		stat_bvhBuildTime = 0L;
		normalFetch.SetCount(0);

		stat_avgMint = 0.;
		stat_avgDirLength = 0.;

		for (int i=0;i<10;i++)
			stat_insideLevel[i] = 0L;
	}

	~BerconMetaball() {
		// Print statistics
		CharStream *out = thread_local(current_stdout);
		out->printf("Metaball statistics\n");
		out->printf("BVH Build time (ms): %d\n", stat_bvhBuildTime);
		out->printf("BVH Branch count: %d\n", bvh->getBrachCount());
		out->printf("BVH Leaf count: %d\n", bvh->getLeafCount());
		out->printf("Rays cast into metaballs: %d\n", stat_RaysCast);
		out->printf("Rays skipped using BVH: %d\n", stat_RaysSkipped);
		out->printf("Ray hits/casts outside: %d / %d\n", stat_HitsOut, stat_RaysOut);
		out->printf("Ray hits/casts inside: %d / %d\n", stat_HitsIn, stat_RaysIn);
		out->printf("Avg bRay depth outside: %f\n", (float)stat_bRayDepthOut/(float)(stat_RaysOut));
		out->printf("Avg bRay depth inside: %f\n", (float)stat_bRayDepthIn/(float)(stat_RaysIn));	

		out->printf("Avg mint: %f\n", stat_avgMint/(double)stat_RaysCast);
		out->printf("Avg dir length: %f\n", stat_avgDirLength/(double)stat_RaysCast);

		for (int i=0;i<10;i++)
			out->printf("Inside %d : %d\n", i, stat_insideLevel[i]);

		for (int i=0;i<normalFetch.Count();i++) {
			out->printf("append myArr (Point3 %f %f %f)\n", normalFetch[i].x, normalFetch[i].y, normalFetch[i].z);
		}

		delete[] blobs;
		delete bvh;
		
	}

	/* VRay interface */

	VR::Vector p;

	int maxIter; int maxNorm; double epsilon; double delta;
	double slice;

	MyRayCache *raycache;	

	int expandable(void) { return false; }
	bool splittable(void) { return true; }
	void getBBox(VR::StaticBox &bbox);
	void split(int dim, VR::real middle, VR::StaticBox &bLeft, VR::StaticBox &bRight);
	int intersect(VR::RSRay &ray);
	
	virtual VR::Vector getGNormal(VR::RSRay &ray);
	virtual VR::Vector getGNormal(VR::RSRayRef& rayref);

	int getMemUsage(void) { return 0; }
	uint32 type(void) { return BERCON_METABALL_PRIM; }

	void init(
		int maxIter, int maxNorm, float epsilon, float delta,
		float cX, float cY, float cZ, float cW, float sX, float sY, float sZ, float sW, float slice,
		VR::Vector* points, int pointCount,
		int simpleSplit,
		VR::GeometryGenerator *owner, int ownerIndex, MyRayCache *rayc);
	//void init(const VR::Vector &center, VR::VR::real radius, int flip, VR::GeometryGenerator *owner, int ownerIndex, MyRayCache *rayc);
};

#endif