/*
	Copyright 2009 Jerry Ylilammi
*/

#include "field.h"

using namespace VR;

/*
	Normal
*/

void FieldValue::addPoints(int* points, int count) {
	Ireal r;
	if (radii)
		for (int i=0;i<count;i++) {			
			r = (pointCloud[points[i]]-location).lengthSqr();
			if (r > radii[points[i]]) continue;
			res += function(radii[points[i]], r);
		}
	else
		for (int i=0;i<count;i++) {
			r = (pointCloud[points[i]]-location).lengthSqr();
			if (r > size2) continue;
			res += function(size2, r);
		}
}

void FieldGrad::addPoints(int* points, int count) {
	Ireal r;
	if (radii)
		for (int i=0;i<count;i++) {			
			for (int j=0;j<4;j++) {
				r = (pointCloud[points[i]]-location[j]).lengthSqr();
				if (r > radii[points[i]]) continue;
				res[j] += function(radii[points[i]], r);
			}
		}
	else
		for (int i=0;i<count;i++) {			
			for (int j=0;j<4;j++) {
				r = (pointCloud[points[i]]-location[j]).lengthSqr();
				if (r > size2) continue;
				res[j] += function(size2, r);
			}
		}	
}

void FieldValueMoving::addPoints(Vector* cloud, int count) {
	Ireal r;
	for (int i=0;i<count;i++) {
		r = (cloud[i]-location).lengthSqr();
		if (r > size2) continue;
		res += function(size2, r);
	}
}

void FieldValueMoving::addPoints(Vector* cloud, Ireal* sizes, VR::Ireal* sizesLin, int count) {
	Ireal r;
	for (int i=0;i<count;i++) {
		r = (cloud[i]-location).lengthSqr();
		if (r > sizes[i]) continue;
		res += function(sizes[i], r);
	}
}

void FieldGradMoving::addPoints(Vector* cloud, int count) {
	Ireal r;
	for (int i=0;i<count;i++) {			
		for (int j=0;j<4;j++) {
			r = (cloud[i]-location[j]).lengthSqr();
			if (r > size2) continue;
			res[j] += function(size2, r);
		}
	}
}

void FieldGradMoving::addPoints(Vector* cloud, Ireal* sizes, VR::Ireal* sizesLin, int count) {
	Ireal r;
	for (int i=0;i<count;i++) {			
		for (int j=0;j<4;j++) {
			r = (cloud[i]-location[j]).lengthSqr();
			if (r > sizes[i]) continue;
			res[j] += function(sizes[i], r);
		}
	}
}

/*
	Average
*/

void FieldValueAvg::addPoints(int* points, int count) {
	Ireal r;
	Ireal w;
	
	if (radii)
		for (int i=0;i<count;i++) {			
			r = (pointCloud[points[i]]-location).lengthSqr();
			Ireal squared = radii[points[i]] * radii[points[i]];
			if (r > squared) continue;
			w = function(squared, r);
			totalP += pointCloud[points[i]] * w;
			totalW += w;				
			totalR += radii[points[i]] * w;
		}
	else
		for (int i=0;i<count;i++) {			
			r = (pointCloud[points[i]]-location).lengthSqr();
			if (r > size2) continue;
			w = function(size2, r);
			totalP += pointCloud[points[i]] * w;
			totalW += w;				
		}
}
   
void FieldGradAvg::addPoints(int* points, int count) {
	Ireal r;
	Ireal w;
	if (radii)		
		for (int i=0;i<count;i++) {			
			for (int j=0;j<4;j++) {
				r = (pointCloud[points[i]]-location[j]).lengthSqr();
				Ireal squared = radii[points[i]] * radii[points[i]];
				if (r > squared) continue;
				w = function(squared, r);
				totalP[j] += pointCloud[points[i]] * w;
				totalW[j] += w;			
				totalR[j] += radii[points[i]] * w;
			}
		}
	else
		for (int i=0;i<count;i++) {			
			for (int j=0;j<4;j++) {
				r = (pointCloud[points[i]]-location[j]).lengthSqr();
				if (r > size2) continue;
				w = function(size2, r);
				totalP[j] += pointCloud[points[i]] * w;
				totalW[j] += w;			
			}
		}
}


void FieldValueMovingAvg::addPoints(Vector* cloud, int count) {
	Ireal r;
	Ireal w;
	for (int i=0;i<count;i++) {			
		r = (cloud[i]-location).lengthSqr();
		if (r > size2) continue;
		w = function(size2, r);
		totalP += cloud[i] * w;
		totalW += w;				
	}
}

void FieldValueMovingAvg::addPoints(Vector* cloud, Ireal* sizes, Ireal* sizesLin, int count) {
	Ireal r;
	Ireal w;
	for (int i=0;i<count;i++) {			
		r = (cloud[i]-location).lengthSqr();
		if (r > sizes[i]) continue;
		w = function(sizes[i], r);
		totalP += cloud[i] * w;
		totalW += w;				
		totalR += sizesLin[i] * w;
	}
}

void FieldGradMovingAvg::addPoints(Vector* cloud, int count) {
	Ireal r;
	Ireal w;
	for (int i=0;i<count;i++) {			
		for (int j=0;j<4;j++) {
			r = (cloud[i]-location[j]).lengthSqr();
			if (r > size2) continue;
			w = function(size2, r);
			totalP[j] += cloud[i] * w;
			totalW[j] += w;			
		}
	}
}

void FieldGradMovingAvg::addPoints(Vector* cloud, Ireal* sizes, Ireal* sizesLin, int count) {
	Ireal r;
	Ireal w;
	for (int i=0;i<count;i++) {			
		for (int j=0;j<4;j++) {
			r = (cloud[i]-location[j]).lengthSqr();
			if (r > sizes[i]) continue;
			w = function(sizes[i], r);
			totalP[j] += cloud[i] * w;
			totalW[j] += w;			
			totalR[j] += sizesLin[i] * w;
		}
	}
}