/*
	Copyright 2009 Jerry Ylilammi
*/

#pragma once

#include "vraybase.h"

using namespace VR;

/*
	The field kernel functions

	Note: size and radius are both squared, taking root is very slow process so we avoid it
*/

// (1. - x^2)^2
inline Ireal fieldFunctionPower2(Ireal size, Ireal radius) {
	Ireal temp = (1. - radius / size);
	return temp * temp;
}

// (1. - x^2)^3
inline Ireal fieldFunctionPower3(Ireal size, Ireal radius) {
	Ireal temp = (1. - radius / size);
	return temp * temp * temp;
}

// (1. - x^2)^4
inline Ireal fieldFunctionPower4(Ireal size, Ireal radius) {
	Ireal temp = (1. - radius / size);
	temp *= temp;
	return temp * temp ;
}

// (1. - x^2)^5
inline Ireal fieldFunctionPower5(Ireal size, Ireal radius) {
	Ireal temp = (1. - radius / size);
	Ireal temp2 = temp * temp;
	temp2 *= temp2;
	return temp * temp2;
}

// (1-x^2)^2*(3-2*(1-x^2))
// 2*x^6-3*x^4+1
inline Ireal fieldFunctionHermit2(Ireal size, Ireal radius) {
	Ireal temp = (1. - radius / size);
	return temp * temp * (3. - 2. * temp);
}

// (1-x^2)^4*(3-2*(1-x^2))
inline Ireal fieldFunctionHermit4(Ireal size, Ireal radius) {
	Ireal temp = (1. - radius / size);
	Ireal temp2 = temp * temp;
	temp2 *= temp2;
	return temp2 * (3. - 2. * temp);
}


// 1.+(17. x^4)/9.-(4. x^6)/9.-(22. x^2)/9.
// 1+17/9*x^4-4/9*x^6-22/9*x^2
// 1+(17*x^4-4*x^6-22*x^2)/9
inline Ireal fieldFunctionWyvill(Ireal size, Ireal radius) {
	Ireal temp = radius / size;	
	Ireal temp2 = temp * temp; 		
	return 1. + 17. / 9. * temp2 - 4.0 / 9. * temp2 * temp - 22. / 9. * temp;	
}