/*
	Copyright 2009 Jerry Ylilammi
*/

#include "instanceMoving.h"
#include "primitive.h"
#include "tomax.h"
#include "misc_ray.h"
#include "maxscrpt/MAXScrpt.h"
#include "readPointCloud.h"

using namespace VR;

BerconMetaballInstanceMoving::BerconMetaballInstanceMoving(BerconMetaball *META, INode *node, VRayCore *vray, int renderID) {
	dynamicType = 1;
	VRenderInstance::init(META, node, vray, renderID);
}

BerconMetaballInstanceMoving::~BerconMetaballInstanceMoving(void) {
}

void BerconMetaballInstanceMoving::renderBegin(TimeValue t, VRayCore *vray) {
	VRenderInstance::renderBegin(t, vray);
	raycache.renderBegin((VR::VRayRenderer*) vray);
}

void BerconMetaballInstanceMoving::renderEnd(VRayCore *vray) {
	VRenderInstance::renderEnd(vray);
	raycache.renderEnd((VR::VRayRenderer*) vray);
}

void BerconMetaballInstanceMoving::frameBegin(TimeValue t, VRayCore *vray) {
	VRenderInstance::frameBegin(t, vray);

	objToWorld=node->GetObjTMAfterWSM(t);
	Matrix3 worldToObj=Inverse(objToWorld);

	camToObj=toMatrix3(toTransform(worldToObj)*vray->getFrameData().camToWorld);
	objToCam=toMatrix3(vray->getFrameData().worldToCam*toTransform(objToWorld));

	tm=toTransform(objToWorld);
	itm=inverse(tm);

	mesh=&dummyMesh;

	BerconMetaball *maxObj=((BerconMetaball*) renderObject);

	MetaballParams params;	
	params.size			= maxObj->pblock2->GetFloat(pb_size, t);
	params.threshold	= maxObj->pblock2->GetFloat(pb_threshold, t);
	params.step			= maxObj->pblock2->GetFloat(pb_step, t);
	params.error		= maxObj->pblock2->GetFloat(pb_error, t);
	params.field		= maxObj->pblock2->GetInt(pb_fieldFunction, t);
	params.useAverage	= (bool)maxObj->pblock2->GetInt(pb_useAverage, t);
	params.cutoff		= maxObj->pblock2->GetFloat(pb_cutoff, t);
	params.trim			= maxObj->pblock2->GetFloat(pb_trim, t);
	params.depth		= maxObj->pblock2->GetInt(pb_depth, t);	
	params.leafLength	= maxObj->pblock2->GetFloat(pb_leafLength, t);
	params.leafSize		= maxObj->pblock2->GetInt(pb_leafSize, t);
	params.texStr		= maxObj->pblock2->GetFloat(pb_texstr, t);
	params.texType		= (bool)maxObj->pblock2->GetInt(pb_textype, t);
	params.clamp		= (bool)maxObj->pblock2->GetInt(pb_clamp, t);
	params.clampTo		= maxObj->pblock2->GetFloat(pb_clampTo, t);	
	params.tex = NULL;
	if (maxObj->pblock2->GetInt(pb_useTexture, t) && params.texStr > 0.0001f) {		
		maxObj->pblock2->GetValue(pb_texmap, t, params.tex, Interval(0,0));
		if (params.tex)
			params.tex->Update(t,Interval(0,0));	
	}
	bool useRadii		= (bool)maxObj->pblock2->GetInt(pb_useScale, t);

	// Check that all parameters are valid
	if (params.clampTo <= params.threshold) params.clampTo = params.threshold + 0.001;

	params.steps		= maxObj->pblock2->GetInt(pb_mbSamples, t) - 1; // Notice -1 correction here

	int* pointCount = new int[params.steps];	
	Vector** points = new Vector*[params.steps*2];
	Ireal** sizes = new Ireal*[params.steps*2];
	for (int i=0; i<params.steps*2; i++) {
		points[i] = NULL;
		sizes[i] = NULL;
	}
	
	int start = (int)(vray->getFrameData().frameStart);
	int end = (int)(vray->getFrameData().frameEnd);
	int stepDur = (float)(end - start) / params.steps;
	for (int i=0; i<params.steps; i++) {				
		pointCount[i] = 0;
		TimeValue time[2]; 
		time[0] = start + (int)(stepDur * i);
		time[1] = start + (int)(stepDur * (i + 1));
		readPointCloud reader;
		reader.defaultRadius = params.size;
		reader.read(useRadii, pointCount[i], points[i*2], points[i*2+1], sizes[i*2], sizes[i*2+1], maxObj->pblock2, time);
	}
	
	raycache.frameBegin((VR::VRayRenderer*) vray);

	META.init(params, points, sizes, pointCount, (GeometryGenerator*) this, ownerIndex, &raycache);
}

void BerconMetaballInstanceMoving::frameEnd(VRayCore *vray) {
	VRenderInstance::frameEnd(vray);
	raycache.frameEnd((VR::VRayRenderer*) vray);
}

void BerconMetaballInstanceMoving::compileGeometry(VRayCore *vray) {
	VR::RayServerInterface *rayserver=vray->getRayServer();
	firstPrimID=-1;
	rayserver->storeMovingPrimitive(static_cast<MovingPrimitive*>(this), static_cast<GeometryGenerator*>(this), rayserver->getNewRenderID());
}

void BerconMetaballInstanceMoving::setIntersectionData(RSRay &rsray, void *isd) {
	IntersectionData &isData=*((IntersectionData*) isd);
	RSIntersection &is=rsray.is;
	BerconMetaballPrimitiveMoving *prim=(BerconMetaballPrimitiveMoving*) rsray.is.primitive;
	isData.primitive=is.primitive;
	isData.sb=(Shadeable*) this;
	isData.sd=(ShadeData*) this;
	isData.si=(ShadeInstance*) this;
	isData.surfaceProps=&surfaceProps;
	isData.volume=NULL;
	isData.skipTag=rsray.is.skipTag;
	isData.faceIndex=0;
	isData.bary.set(0.0f, 0.0f, 0.0f);
	isData.wpoint=rsray.p+TracePoint(rsray.dir)*is.t;
	isData.wpointCoeff=is.t;
	isData.gnormal=prim->getGNormal(rsray);
	isData.normal=isData.gnormal;
}

void BerconMetaballInstanceMoving::setIntersectionData(RSRayRef& rsrayref, void *isd) {
	IntersectionData &isData=*((IntersectionData*) isd);
	BerconMetaballPrimitiveMoving *prim=(BerconMetaballPrimitiveMoving*) rsrayref.getGenericPrimitive();
	isData.primitive = rsrayref.getGenericPrimitive();
	isData.sb=(Shadeable*) this;
	isData.sd=(ShadeData*) this;
	isData.si=(ShadeInstance*) this;
	isData.surfaceProps=&surfaceProps;
	isData.volume=NULL;
	isData.skipTag = rsrayref.getSkipTag();
	isData.faceIndex=0;
	isData.bary.set(0.0f, 0.0f, 0.0f);
	isData.wpoint= rsrayref.getOrigin() + TracePoint(rsrayref.getDirection()) * rsrayref.getDistance();
	isData.wpointCoeff = rsrayref.getDistance();
	isData.gnormal= prim->getGNormal(rsrayref);
	isData.normal= isData.gnormal;
}

// From RenderInstance
Interval BerconMetaballInstanceMoving::MeshValidity(void) { return FOREVER; }
Point3 BerconMetaballInstanceMoving::GetFaceNormal(int faceNum) { return Point3(0,0,1); }
Point3 BerconMetaballInstanceMoving::GetFaceVertNormal(int faceNum, int vertNum) { return Point3(0,0,1); }
void BerconMetaballInstanceMoving::GetFaceVertNormals(int faceNum, Point3 n[3]) {}
Point3 BerconMetaballInstanceMoving::GetCamVert(int vertNum) { return Point3(0,0,0); }
void BerconMetaballInstanceMoving::GetObjVerts(int fnum, Point3 obp[3]) {}
void BerconMetaballInstanceMoving::GetCamVerts(int fnum, Point3 cp[3]) {}

// From Shadeable
void BerconMetaballInstanceMoving::shade(VR::VRayContext &vri) {VRenderInstance::fullShade((VR::VRayInterface&) vri);}

// From ShadeData
VR::Vector BerconMetaballInstanceMoving::getUVWcoords(const VR::VRayContext &vri, int channel) {return VR::Vector(0,0,0);}
void BerconMetaballInstanceMoving::getUVWderivs(const VR::VRayContext &vri, int channel, VR::Vector derivs[2]) {
		derivs[0]=normalize(tm.m[0]);derivs[1]=normalize(tm.m[1]);}		
void BerconMetaballInstanceMoving::getUVWbases(const VR::VRayContext &vri, int channel, VR::Vector bases[3]) {
		bases[0]=normalize(tm.m[0]);bases[1]=normalize(tm.m[1]);bases[2]=normalize(tm.m[2]);}
VR::Vector BerconMetaballInstanceMoving::getUVWnormal(const VR::VRayContext &vri, int channel) {return normalize(tm.m[2]);}
int BerconMetaballInstanceMoving::getMtlID(const VR::VRayContext &vri) { return 0; }
int BerconMetaballInstanceMoving::getGBufID(void) { return 0; }
int BerconMetaballInstanceMoving::getSurfaceRenderID(const VR::VRayContext &rc) { return VRenderInstance::renderID; }
int BerconMetaballInstanceMoving::getMaterialRenderID(const VR::VRayContext &rc) { return 0; }
void BerconMetaballInstanceMoving::getSmoothUVWbases(const VR::VRayContext &vri, int channel, VR::Vector bases[3]) {getUVWbases(vri, channel, bases);}

// From ShadeInstance
VR::Vector BerconMetaballInstanceMoving::getBasePt(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.wpoint; }
VR::Vector BerconMetaballInstanceMoving::getBasePtObj(const VR::VRayContext &vri) { return vri.rayresult.wpoint; }
VR::Vector BerconMetaballInstanceMoving::worldToObjectVec(const VR::VRayContext &vri, const VR::Vector &d) { return d; }
VR::Vector BerconMetaballInstanceMoving::worldToObjectPt(const VR::VRayContext &vri, const VR::Vector &p) { return p; }
VR::Vector BerconMetaballInstanceMoving::objectToWorldVec(const VR::VRayContext &vri, const VR::Vector &d) { return d; }
VR::Vector BerconMetaballInstanceMoving::objectToWorldPt(const VR::VRayContext &vri, const VR::Vector &p) { return p; }
VR::Vector BerconMetaballInstanceMoving::getBaseGNormal(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.gnormal; }
VR::Vector BerconMetaballInstanceMoving::getBaseNormal(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.origNormal; }
VR::Vector BerconMetaballInstanceMoving::getBaseBumpNormal(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.normal; }
VR::Vector BerconMetaballInstanceMoving::getShadowPt(const VR::VRayContext &vri) { return vri.rayresult.wpoint; }
VR::Vector BerconMetaballInstanceMoving::getVelocity(const VR::VRayContext &vri) { return VR::Vector(0,0,0); }

// From DynamicPrimitive
int BerconMetaballInstanceMoving::expandable(void) { return true; }
bool BerconMetaballInstanceMoving::splittable(void) { return true; }