/*
	Copyright 2009 Jerry Ylilammi
*/

#pragma once

#ifndef __PRIMITIVE_H__
#define __PRIMITIVE_H__

#include "BVHStatic.h"
#include "BVHMoving.h"
#include "BerconShadeContext.h"

#define BERCON_METABALL_PRIM_STATIC 0x10011256
#define BERCON_METABALL_PRIM_MOVING 0x10011257

#define NORMALDELTA 0.001f // Delta used in the normal computations below

inline void computeNormalLocations(VR::Vector* location, VR::Vector pos) {
	location[0] = pos;
	location[1] = pos - VR::Vector(NORMALDELTA,0.f,0.f);
	location[2] = pos - VR::Vector(0.f,NORMALDELTA,0.f);
	location[3] = pos - VR::Vector(0.f,0.f,NORMALDELTA);
}

inline void computeNormalResult(VR::Ireal* res) {
	res[1] = res[0] - res[1];
	res[2] = res[0] - res[2];
	res[3] = res[0] - res[3];
	VR::Ireal length = sqrt(res[1]*res[1] + res[2]*res[2] + res[3]*res[3]);
	res[1] /= length;
	res[2] /= length;
	res[3] /= length;
}

struct MyCacheStruct {	
	VR::Vector rO;	
};

struct MetaballParams {
	float size;
	float threshold;
	float step;
	float error;
	int field;
	bool useAverage;
	float cutoff;
	float trim;
	int depth;
	float leafLength;
	float leafSize;
	int steps;
	Texmap *tex;
	bool texType;
	float texStr;
	bool clamp;
	float clampTo;
	//Matrix3 tm;
	//Matrix3 itm;
};

typedef VR::RayCache<MyCacheStruct> MyRayCache;

/*inline float smoothstep(float d, float low, float high) { 
	d = (d-low)/(high-low);
	if (d < 0) return 0.f;
	if (d > 1) return 1.f;
	return (d*d*(3.f-2.f*d));
}*/

inline VR::Ireal smoothstep1(VR::Ireal d, VR::Ireal low) {
	d = (d - low) / (1.0 - low);
	if (d < 0) return 0.0;
	if (d > 1) return 1.0;
	return (d * d * (3.0 - 2.0 * d));
}

/*
	Common stuff between static and moving versions of the metaball primitive
	The primitive does the actual interscetion part
*/
class  BerconMetaballPrimitive {
public:
	// Parameters	
	VR::Ireal blobSize;	
	VR::Ireal blobThreshold;
	VR::Ireal stepLength;
	VR::Ireal maxError;
	VR::Ireal (*function)(VR::Ireal size2, VR::Ireal squaredDistance);
	bool useAverage;	
	VR::Ireal cutoff;
	VR::Ireal trim;

	Texmap *tex;
	bool texType;
	VR::real texStr;
	VR::real texStr1; // 1.f - texStr

	bool clamp;
	VR::Ireal clampTo;
	//Matrix3 tm;
	//Matrix3 itm;

	// Generated values
	VR::Ireal blobSize2;

	// Cache
	MyRayCache *raycache;
	
	// Methods
	inline void applyTexmap(VR::Vector pos, VR::Ireal& res) {
		if (tex) {
			BerconShadeContext sc;
			sc.Init(Point3(pos.x, pos.y, pos.z));			
			float textureVal = tex->EvalMono(sc);
			if (clamp && res > clampTo) res = clampTo;
			if (texType) res -= textureVal * texStr; else res *= texStr1 + textureVal * texStr;
		}
	}
	
	inline void applyTexmap(BerconShadeContext& sc, VR::Vector pos, VR::Ireal& res) {
		sc.Init(Point3(pos.x, pos.y, pos.z));			
		float textureVal = tex->EvalMono(sc);
		if (clamp && res > clampTo) res = clampTo;
		if (texType) res -= textureVal * texStr; else res *= texStr1 + textureVal * texStr;
	}
	
	inline void applyTexmap(VR::Ireal textureVal, VR::Ireal& res) {
		if (clamp && res > clampTo) res = clampTo;
		if (texType) res -= textureVal; else res *= textureVal; // texStr has been precomputed in getTexmap()
	}
	
	inline VR::Ireal getTexmap(BerconShadeContext& sc, VR::Vector pos) {
		sc.Init(Point3(pos.x, pos.y, pos.z));			
		if (texType)
			return tex->EvalMono(sc) * texStr;
		return texStr1 + tex->EvalMono(sc) * texStr;
	}

	inline bool testThreshold(VR::Ireal f, bool insideField) { return insideField ? f < blobThreshold : f > blobThreshold; }
};

#endif