/*
	Copyright 2009 Jerry Ylilammi
*/

#pragma once

#include "primitive.h"

/*
	Static version of the primitive
*/
class BerconMetaballPrimitiveStatic: public BerconMetaballPrimitive, public VR::StaticPrimitive {
private:			
	VR::StaticBox bb;
	int blobsCount;
	VR::Vector *blobs;
	VR::Ireal *radii;
	BoundingVolumeHierarchy* bvh;
	
public:
	BerconMetaballPrimitiveStatic() {
		bvh = new BoundingVolumeHierarchy();
	}

	~BerconMetaballPrimitiveStatic() {
		delete[] blobs;
		delete[] radii;
		delete bvh;	
	}

	bool intersectTraditional(VR::Vector rO, VR::Vector rD, BVHStep* first, bool insideField, VR::Ireal stepLength, VR::Ireal maxError, VR::Ireal &dist);
	bool intersectAverage(VR::Vector rO, VR::Vector rD, BVHStep* first, bool insideField, VR::Ireal stepLength, VR::Ireal maxError, VR::Ireal &dist);	
	VR::Vector normal(VR::Vector location);

	inline void getAverageResult(FieldValueAvg &field, VR::Vector &pos);
	inline void getAverageResult(FieldGradAvg &grad);

	/* VRay interface */
	int expandable(void) { return false; }
	bool splittable(void) { return true; }	
	int intersect(VR::RSRay &ray);
	virtual VR::Vector getGNormal(VR::RSRay &ray);
	virtual VR::Vector getGNormal(VR::RSRayRef& rayref);
	int getMemUsage(void) { return 0; }

	uint32 type(void) { return BERCON_METABALL_PRIM_STATIC; }
	void getBBox(VR::StaticBox &bbox);
	void split(int dim, VR::real middle, VR::StaticBox &bLeft, VR::StaticBox &bRight);
	void init(MetaballParams params, VR::Vector* points, VR::Ireal *sizes, int pointCount, VR::GeometryGenerator *owner, int ownerIndex, MyRayCache *rayc);
};